{****************************************************************************
 *             } Unit UKBD;{ version 1.2 (c) 1999 Michal Cihar              *
 ****************************************************************************
 * Contact: WWW: http://miso.fd.cz                                          *
 *          e-mail: ciharm@fel.cvut.cz                                      *
 ****************************************************************************
 * History                                                                  *
 * -------                                                                  *
 * Version 1.1 news: converted to assembler                                *
 * Version 1.2 news: new function whatkeypressed                           *
 *                   new status functions: enhanced_keyboard, LED-states,  *
 *                    right ALT and CTRL                                    *
 *                   new functions getkbdID/getkbdIDstring                 *
 *                   better documentation                                  *
 *                   CRTL-BREAK interrupt handler - sets variable          *
 *                    CtrlBreakHit and you can define own procedure, that   *
 *                    is called while CRTL-BREAK is pressed                 *
 *                   new function readscan - reads scan-code               *
 *                   scan-code constants                                   *
 *                   keyboard interrupt handler - allows to ignore some    *
 *                    scan-codes                                            *
 *                   ^ignores keys in set forgetkeys or accepts keys in    *
 *                    set onlykeys - can be set via constants s_keyname     *
 *                                                                          *
 *                                                                          *
 ****************************************************************************}
{{$DEFINE SETINT} {if this defined unit allocates interrupts}

interface
const UnitID='UKbd';
      UnitCopyright='1999 Michal Cihar';
      UnitVersion='1.2';
      UnitRelease='27.1.1999';

{keyboard and buffer ckear, reset,...}
{clears buffer:}
procedure clearbuffer;
{resets keyboard and set default settings (my :-):}
procedure init_kbd;
{resets keyboard (the same procedure as during computer inicialization):}
procedure reset_kbd;
{sleeps keyboard (no reaction on any key):}
procedure sleep_kbd;
{wakes keyboard - use after slepp_kbd:}
procedure wake_kbd;

{key read, write functions:}
{true if key was pressed:}
function keypressed : boolean;
{0 if no key was pressed, otherwise ASCII value - key is not removed from
buffer:}
function whatkeypressed : byte;
{reads key, returns high byte - scancode, low byte - asciicode:}
function read_key : word;
{reads key, returns character (#0 if enhanced, next call enhanced ASCII):}
function readkey:char;
{reads key, returns scancode:}
function readscan : byte;
{writes char into buffer (returns true if succesful):}
function tobuffer(const character:char):boolean;

{num, caps, scroll locks, LEDs, ...}
{set LEDs (light emmision diode) - note that this sets ONLY LEDs:}
procedure set_led(caps,num,scroll:boolean);
{next 3 functions read LED state - againg ONLY LED:}
function scroll_lock_LED:boolean;
function num_lock_LED:boolean;
function caps_lock_LED:boolean;
{next 3 functions set lock value - lock and LED:}
procedure setscroll_lock(value:boolean);
procedure setnum_lock(value:boolean);
procedure setcaps_lock(value:boolean);
{next 3 functions read lock value - lock:}
function scroll_lock:boolean;
function num_lock:boolean;
function caps_lock:boolean;
{next 3 functions determines whether is pressed:}
function scroll_lock_pressed:boolean;
function num_lock_pressed:boolean;
function caps_lock_pressed:boolean;

{insert functions:}
{sets insert status:}
procedure setinsert_status(value:boolean);
{reads insert status:}
function insert_status:boolean;
{determines whether insert is pressed:}
function insert_pressed:boolean;

{folowing functions return info about SHIFT, ALT and CTRL keys, if true key
is pressed, each is without prefix, L_ and R_ that means: no matter which,
left and right.
Note that some keyboards/BIOSes don't set ALT properly - sometimes on place
where should be ALT value is left ALT value.
Note2: some combination of these keys pressed disables keyboard/BIOS to set
some other flags - two keys are allways OK, more should have some problems.}
function shift:boolean;
function L_shift:boolean;
function R_shift:boolean;
function ctrl:boolean;
function L_ctrl:boolean;
function R_ctrl:boolean;
function alt:boolean;
function L_alt:boolean;
function R_alt:boolean;

{true if print screen key is pressed:}
function printscreen_pressed:boolean;
{true if pause state is active:}
function pause_status:boolean;
{true if enhanced keyboard is available:}
function enhanced_keyboard:boolean;

{WARNING for functions getkbdID and getkbdIDstring:
These are not tested, whether they return correct value, because I didn't
find any keyboard supporting this function (specially for Japan keyboards.)
So take results carefullly and report me any bugs.}
{returns keyboard ID or FFFFh if not supported:}
function getkbdID:word;
{returns string according to getkbdID:}
function getkbdIDstring:string;

{$IFDEF SETINT}
{CTRL-BREAK handling routines}
{if Ctrl-Break is pressed is true - false is never set by unit:}
const CtrlBreakHit: boolean = false;
{set CTRL-BREAK interrupt handler to that one while starting unit:}
procedure setoldctrlbreak;
{set CTRL-BREAK interrupt handler to unit UKbdctrlbreak procedure:}
procedure setUKbdctrlbreak;
{procedure that can be called from UKbd CTRL-BREAK interrupt handler:
Note: the procedure sould be far}
var userctrlbreak:procedure;
{after using ^ when you don't need it then set it to dummy:}
procedure dummy;
{$ENDIF}

{scan-codes constants:}
const _space=57;
      _escape=1;
      _enter=28;
      _tab=15;
      {numlock keys:}
      num_0=82;
      num_1=79;
      num_2=80;
      num_3=81;
      num_4=75;
      num_5=76;
      num_6=77;
      num_7=71;
      num_8=72;
      num_9=73;
      _F1=59;
      _F2=60;
      _F3=61;
      _F4=62;
      _F5=63;
      _F6=64;
      _F7=65;
      _F8=66;
      _F9=67;
      _F10=68;
      _F11=87;
      _F12=88;
      _a=30;
      _n=44;
      _y=27;
      _r=14;
      _f=33;
{scan-codes constants (first is press second release) for use with
forgetkeys or onlykeys sets:}
{
      s:set of byte=[,+127];
}
const s_enter:set of byte=[_enter,_enter+127];
      s_space:set of byte=[_space,_space+127];
      s_escape:set of byte=[_escape,_escape+127];
      s_a:set of byte=[_a,_a+127];
      s_n:set of byte=[_n,_n+127];
      s_y:set of byte=[_y,_y+127];
      s_r:set of byte=[_r,_r+127];
      s_f:set of byte=[_f,_f+127];


{these sets are used by UKbd keyboard interrupt hadler to:
forgetkeys - ignore scan-codes that are in set
onlykeys - if this is not empty only scancodes in this set are passed to
stadandart keyboard interrupt handler.
Note: please insert all values into these sets in pairs: value, value+127
this ensures non-error run of your program. You can use some predefined
set-constants (s_keyname - see upper) and pascal procedures include and/or
exclude.
Note2: if only key is NOT empty, then only scan-codes in onlykeys set are
passed to stadandart keyboard interrupt handler.}
const forgetkeys:set of byte=[];
      onlykeys:set of byte=[];

implementation

uses dos;
var lastext:char;
    stav1:byte absolute 0000:$0417; {looks ugly? :-}
    stav2:byte absolute 0000:$0418;
    info1:byte absolute 0000:$0496;
    info2:byte absolute 0000:$0497;
    support:byte; {from INT 16h/AH=09h - support of functions}
    r:registers;
{$IFDEF SETINT}
    oldctrlbreak:procedure; {stored old ctr-break interrupt}
    oldkeyboard:procedure;  {stored old keyboard interrupt}
    oldexit : pointer;      {stored exit procedure}
{$ENDIF}
const emptyset:set of byte=[];

procedure dummy;
begin;
end;

{$IFDEF SETINT}
procedure UKbdctrlbreak;interrupt;assembler;
asm
  mov CtrlBreakHit, 1
  call userctrlbreak
end;

procedure UKbdkeyboard;interrupt;
begin;
if onlykeys<>emptyset then
begin;
  if not(port[$60] in onlykeys) then
  asm
    mov al,$20
    out $20,al
  end else
  begin;
    inline($9c);
    oldkeyboard;
  end;
end else
begin;
  if port[$60] in forgetkeys then
  asm
    mov al,$20
    out $20,al
  end else
  begin;
    inline($9c);
    oldkeyboard;
  end;
end;
end;
{$ENDIF}

procedure init_kbd;assembler;
asm
  mov al,0ffh
  out 60h,al
  mov ax, 0305h
  mov bx,0
  int 16h
end;

function getkbdID:word;assembler;
asm
  mov support, ah
  and ah, 16
  cmp ah, 16
  jne @not
  mov ah, 0Ah
  int 16h
  mov ax,bx
  jmp @end
@not:
  mov ax, 0ffffh
@end:
end;

function getkbdIDstring:string;
begin;
case getkbdID of 0:getkbdIDstring:='not connected';
             $ffff:getkbdIDstring:='detection not supported';
             $41ab:getkbdIDstring:='Japan "G" or MF2 keyboard in translate mode';
             $54ab:getkbdIDstring:='Japan "P" keyboard in translate mode';
             $83ab:getkbdIDstring:='Japan "G" or MF2 keyboard in pass-throught mode';
             $84ab:getkbdIDstring:='Japan "P" keyboard in pass-throught mode';
             $90ab:getkbdIDstring:='Japan "G" old model';
             $91ab:getkbdIDstring:='Japan "P" old model';
             $92ab:getkbdIDstring:='Japan "A" old model';
             end;
end;

procedure set_led(caps,num,scroll:boolean); assembler;
asm
  {signal change LED}
  mov al,0edh
  out 60h,al

  {delay 5000 microseconds}
  mov cx, 0
  mov dx, 5000
  mov ah, 86h
  int 15h

  {calculate data}
  xor al, al
  cmp caps,1
  jne @01
  or al, 4
@01:
  cmp num,1
  jne @02
  or al, 2
@02:
  cmp scroll,1
  jne @03
  or al, 1
@03:
  {send data}
  out 60h,al
end;

procedure reset_kbd;assembler;
 asm
  mov al,0ffh
  out 60h,al
 end;

procedure sleep_kbd;assembler;
 asm
  mov al,0f5h
  out 60h,al
 end;

procedure wake_kbd;assembler;
 asm
  mov al,0f4h
  out 60h,al
 end;

function keypressed : boolean;assembler;
asm
   mov  ax, 40h
   mov  es, ax
   mov  dx, es:[1ah]
   mov  bx, es:[1ch]
   xor  ax, ax
   cmp  dx, bx
   je  @end
   mov  al, 1
 @end:
end;

function whatkeypressed : byte;assembler;
asm
   mov  ax, 40h
   mov  es, ax
   mov  bx, es:[1ah]
   mov  dx, es:[1ch]
   xor  ax, ax
   cmp  dx, bx
   je  @end
   mov al, es:[bx]
 @end:
end;

procedure clearbuffer;assembler;
asm
   mov  ax, 40h
   mov  es, ax
   mov  ax, es:[1ch]
   mov  es:[1ah], ax
end;

function read_key : word;assembler;
asm
   mov  ax, 1000h
   int  16h
end;

function readscan : byte;assembler;
asm
   mov  ax, 1000h
   int  16h
   mov al, ah
end;

function readkey:char; assembler;
asm;
   cmp lastext, 0
   jne @01
   mov ah, 00h
   int 16h
   cmp al, 0
   jne @end
   mov lastext, ah
   jmp @end
@01:
   mov al, lastext
   mov lastext, 0
@end:
end;

function tobuffer(const character:char):boolean; assembler;
asm
 mov ah, 05h
 mov cl, character
 mov ch, 0
 int 16h
 mov ah, 0
 cmp al, ah
 jne @end
 mov  al, 1
@end:
end;

{ This is fully functionable, but a bit long :-)
function L_shift:boolean;assembler;
asm
 mov ax, 040h
 mov es, ax
 mov al, es:017h
 and al,2
 cmp al,2
 je @1
 mov al, 0
 jmp @end
@1:
 mov al, 1
@end:
end;}


function L_shift:boolean;
begin;
 l_shift:=(stav1 and 2)=2;
end;

function R_shift:boolean;
begin;
 r_shift:=(stav1 and 1)=1;
end;

function shift:boolean;
begin;
 shift:=((stav1 and 2)=2)or((stav1 and 1)=1);
end;

function ctrl:boolean;
begin;
 ctrl:=(stav1 and 4)=4;
end;

function alt:boolean;
begin;
 alt:=(stav1 and 8)=8;
end;

function scroll_lock:boolean;
begin;
 scroll_lock:=(stav1 and 16)=16;
end;

function num_lock:boolean;
begin;
 num_lock:=(stav1 and 32)=32;
end;

function caps_lock:boolean;
begin;
 caps_lock:=(stav1 and 64)=64;
end;

function insert_status:boolean;
begin;
 insert_status:=(stav1 and 128)=128;
end;

function L_ctrl:boolean;
begin;
 l_ctrl:=(stav2 and 1)=1;
end;

function L_alt:boolean;
begin;
 l_alt:=(stav2 and 2)=2;
end;

function printscreen_pressed:boolean;
begin;
 printscreen_pressed:=(stav2 and 4)=4;
end;

function pause_status:boolean;
begin;
 pause_status:=(stav2 and 8)=8;
end;

function scroll_lock_pressed:boolean;
begin;
 scroll_lock_pressed:=(stav2 and 16)=16;
end;

function num_lock_pressed:boolean;
begin;
 num_lock_pressed:=(stav2 and 32)=32;
end;

function caps_lock_pressed:boolean;
begin;
 caps_lock_pressed:=(stav2 and 64)=64;
end;

function insert_pressed:boolean;
begin;
 insert_pressed:=(stav2 and 128)=128;
end;

procedure setscroll_lock(value:boolean);
begin;
 if not(((stav1 and 16)=16)=value) then
    if value then stav1:=stav1+16 else stav1:=stav1-16;
 if value then set_led(caps_lock,num_lock,true) else set_led(caps_lock,num_lock,false);
end;

procedure setnum_lock(value:boolean);
begin;
 if not(((stav1 and 32)=32)=value) then
    if value then stav1:=stav1+32 else stav1:=stav1-32;
 if value then set_led(caps_lock,true,scroll_lock) else set_led(caps_lock,false,scroll_lock);
end;

procedure setcaps_lock(value:boolean);
begin;
 if not(((stav1 and 64)=64)=value) then
    if value then stav1:=stav1+64 else stav1:=stav1-64;
 if value then set_led(true,num_lock,scroll_lock) else set_led(false,num_lock,scroll_lock);
end;

procedure setinsert_status(value:boolean);
begin;
 if not(((stav1 and 125)=128)=value) then
    if value then stav1:=stav1+128 else stav1:=stav1-128;
end;

function enhanced_keyboard:boolean;
begin;
 enhanced_keyboard:=(INFO1 and 16)=16;
end;

function R_ctrl:boolean;
begin;
 R_ctrl:=(INFO1 and 4)=4;
end;

function R_alt:boolean;
begin;
 R_alt:=(INFO1 and 8)=8;
end;

function scroll_lock_LED:boolean;
begin;
 scroll_lock_LED:=(INFO2 and 1)=1;
end;

function num_lock_LED:boolean;
begin;
 num_lock_LED:=(INFO2 and 2)=2;
end;

function caps_lock_LED:boolean;
begin;
 caps_lock_LED:=(INFO2 and 4)=4;
end;

{$IFDEF SETINT}
procedure setoldctrlbreak;
begin;
setintvec($1b,@oldctrlbreak);
end;

procedure setUKbdctrlbreak;
begin;
setintvec($1b,@UKbdctrlbreak);
end;

procedure exit_ukbd;far;
begin
  exitproc := oldexit;
  setintvec($1b,@oldctrlbreak);
  setintvec($9,@oldkeyboard);
end;
{$ENDIF}


begin;
{$IFDEF SETINT}
oldexit  := exitproc;
exitproc := @exit_ukbd;
{$ENDIF}
asm
mov lastext,0
mov ah, 09h
int 16h
mov support, al
end;
{$IFDEF SETINT}
userctrlbreak:=dummy;
getintvec($1b,@oldctrlbreak);
setintvec($1b,@UKbdctrlbreak);
getintvec($9,@oldkeyboard);
setintvec($9,@ukbdkeyboard);}
{$ENDIF}
end.